import skimage.color
import skimage.segmentation
import skimage.filters
import skimage.morphology
import skimage.util
import skimage.transform
import skimage.measure
import skimage.io

from serpent.utilities import is_unix, is_windows
from serpent.utilities import SerpentError

try:
    if is_unix():
        import tesserocr
    elif is_windows():
        import pytesseract
except ImportError:
    raise SerpentError("Setup has not been been performed for the OCR module. Please run 'serpent setup ocr'")

import editdistance

from PIL import Image


def locate_string(query_string, image, fuzziness=0, ocr_preset=None, offset_x=0, offset_y=0):
    images, text_regions = extract_ocr_candidates(
        image,
        gradient_size=ocr_preset["extract"]["gradient_size"],
        closing_size=ocr_preset["extract"]["closing_size"]
    )

    detected_strings = list()

    for image in images:
        detected_strings.append(
            perform_ocr(
                image,
                scale=ocr_preset["perform"]["scale"],
                order=ocr_preset["perform"]["order"],
                horizontal_closing=ocr_preset["perform"]["horizontal_closing"],
                vertical_closing=ocr_preset["perform"]["vertical_closing"]
            )
        )

    if query_string in detected_strings:
        text_region = list(text_regions[detected_strings.index(query_string)])

        text_region[0] += offset_y
        text_region[1] += offset_x
        text_region[2] += offset_y
        text_region[3] += offset_x

        return text_region

    if len(detected_strings):
        edit_distances = [editdistance.eval(query_string, string) for string in detected_strings]
        minimum_edit_distance = min(edit_distances)

        if minimum_edit_distance <= fuzziness:
            text_region = list(text_regions[edit_distances.index(minimum_edit_distance)])

            text_region[0] += offset_y
            text_region[1] += offset_x
            text_region[2] += offset_y
            text_region[3] += offset_x

            return text_region

    return None


def extract_ocr_candidates(image, gradient_size=3, closing_size=10, minimum_area=100, minimum_aspect_ratio=2):
    gray_image = skimage.color.rgb2gray(image)
    gradient_image = skimage.filters.rank.gradient(gray_image, skimage.morphology.disk(gradient_size))
    bw_image = gradient_image > skimage.filters.threshold_otsu(gradient_image)
    bw_image = skimage.morphology.closing(bw_image, skimage.morphology.rectangle(1, closing_size))
    label_image = skimage.measure.label(bw_image)

    regions = skimage.measure.regionprops(label_image)

    text_regions = list()

    for region in regions:
        if region.area > minimum_area:
            y0, x0, y1, x1 = region.bbox
            region_image = bw_image[y0:y1, x0:x1]

            aspect_ratio = region_image.shape[1] / region_image.shape[0]

            black_pixel_count = region_image[region_image == 0].size
            white_pixel_count = region_image[region_image == 1].size

            if aspect_ratio >= minimum_aspect_ratio and white_pixel_count > black_pixel_count and y1 - y0 >= 8:
                text_regions.append(region.bbox)

    ocr_candidates = list()

    for text_region in text_regions:
        ocr_candidates.append(gray_image[text_region[0]:text_region[2], text_region[1]:text_region[3]])

    return ocr_candidates, text_regions


def perform_ocr(image, scale=10, order=5, horizontal_closing=10, vertical_closing=5):
    image = skimage.transform.resize(
        image,
        (image.shape[0] * scale, image.shape[1] * scale),
        mode="edge",
        order=order
    )

    image = image > skimage.filters.threshold_otsu(image)

    black_pixel_count = image[image == 0].size
    white_pixel_count = image[image == 1].size

    if black_pixel_count > white_pixel_count:
        image = skimage.util.invert(image)

    image = skimage.morphology.closing(image, skimage.morphology.rectangle(1, horizontal_closing))
    image = skimage.morphology.closing(image, skimage.morphology.rectangle(vertical_closing, 1))

    image = skimage.util.img_as_ubyte(image)

    if is_unix():
        return tesserocr.image_to_text(
            Image.fromarray(image),
            psm=tesserocr.PSM.SINGLE_LINE,
            oem=tesserocr.OEM.TESSERACT_ONLY
        ).strip()
    elif is_windows():
        return pytesseract.image_to_string(Image.fromarray(image))



from serpent.machine_learning.context_classification.context_classifier import ContextClassifier

from serpent.utilities import SerpentError

try:
    from keras.preprocessing.image import ImageDataGenerator
    from keras.applications.inception_v3 import InceptionV3, preprocess_input
    from keras.layers import Dense, GlobalAveragePooling2D
    from keras.models import Model, load_model
    from keras.callbacks import ModelCheckpoint
except ImportError:
    raise SerpentError("Setup has not been been performed for the ML module. Please run 'serpent setup ml'")

import skimage.transform

import numpy as np

import serpent.cv


class CNNInceptionV3ContextClassifier(ContextClassifier):

    def __init__(self, input_shape=None):
        super().__init__()
        self.input_shape = input_shape

        self.training_generator = None
        self.validation_generator = None

    def train(self, epochs=3, autosave=False, validate=True):
        if validate and (self.training_generator is None or self.validation_generator is None):
            self.prepare_generators()

        base_model = InceptionV3(
            weights="imagenet",
            include_top=False,
            input_shape=self.input_shape
        )

        output = base_model.output
        output = GlobalAveragePooling2D()(output)
        output = Dense(1024, activation='relu')(output)

        predictions = Dense(len(self.training_generator.class_indices), activation='softmax')(output)
        self.classifier = Model(inputs=base_model.input, outputs=predictions)

        for layer in base_model.layers:
            layer.trainable = False

        self.classifier.compile(
            optimizer="rmsprop",
            loss="categorical_crossentropy",
            metrics=["accuracy"]
        )

        callbacks = []

        if autosave:
            callbacks.append(ModelCheckpoint(
                "datasets/context_classifier_{epoch:02d}-{val_loss:.2f}.model",
                monitor='val_loss',
                verbose=0,
                save_best_only=False,
                save_weights_only=False,
                mode='auto',
                period=1
            ))

        self.classifier.fit_generator(
            self.training_generator,
            samples_per_epoch=self.training_sample_count,
            nb_epoch=epochs,
            validation_data=self.validation_generator,
            nb_val_samples=self.validation_sample_count,
            class_weight="auto",
            callbacks=callbacks
        )

    def validate(self):
        pass

    def predict(self, input_frame):
        if self.training_generator is None or self.validation_generator is None:
            self.prepare_generators()

        source_min = 0

        if str(input_frame.dtype) == "uint8":
            source_max = 255
        elif str(input_frame.dtype) == "float64":
            source_max = 1

        input_frame = np.array(serpent.cv.normalize(
            input_frame,
            source_min,
            source_max,
            target_min=-1,
            target_max=1
        ), dtype="float32")

        class_mapping = self.training_generator.class_indices
        class_probabilities = self.classifier.predict(input_frame[None, :, :, :])[0]

        max_probability_index = np.argmax(class_probabilities)
        max_probability = class_probabilities[max_probability_index]

        if max_probability < 0.5:
            return None

        for class_name, i in class_mapping.items():
            if i == max_probability_index:
                return class_name

    def save_classifier(self, file_path):
        if self.classifier is not None:
            self.classifier.save(file_path)

    def load_classifier(self, file_path):
        self.classifier = load_model(file_path)

    def prepare_generators(self):
        training_data_generator = ImageDataGenerator(preprocessing_function=preprocess_input)
        validation_data_generator = ImageDataGenerator(preprocessing_function=preprocess_input)

        self.training_generator = training_data_generator.flow_from_directory(
            "datasets/current/training",
            target_size=(self.input_shape[0], self.input_shape[1]),
            batch_size=32
        )

        self.validation_generator = validation_data_generator.flow_from_directory(
            "datasets/current/validation",
            target_size=(self.input_shape[0], self.input_shape[1]),
            batch_size=32
        )



from serpent.machine_learning.reinforcement_learning.replay_memory import ReplayMemory
from serpent.machine_learning.reinforcement_learning.epsilon_greedy_q_policy import EpsilonGreedyQPolicy

from serpent.visual_debugger.visual_debugger import VisualDebugger

from serpent.utilities import SerpentError

try:
    from keras.models import Model
    from keras.layers import Dense, Flatten, Convolution2D, MaxPooling2D, AveragePooling2D, Input, merge
    from keras.optimizers import Adam, rmsprop
except ImportError:
    raise SerpentError("Setup has not been been performed for the ML module. Please run 'serpent setup ml'")

import numpy as np

import random
import itertools


class DQN:

    def __init__(
        self,
        input_shape=None,
        input_mapping=None,
        replay_memory_size=10000,
        batch_size=32,
        action_space=None,
        max_steps=1000000,
        observe_steps=None,
        initial_epsilon=1.0,
        final_epsilon=0.1,
        gamma=0.99,
        model_file_path=None,
        model_learning_rate=2.5e-4,
        override_epsilon=False
    ):
        self.type = "DQN"
        self.input_shape = input_shape
        self.replay_memory = ReplayMemory(memory_size=replay_memory_size)
        self.batch_size = batch_size
        self.action_space = action_space
        self.action_count = len(self.action_space.combinations)
        self.action_input_mapping = self._generate_action_space_combination_input_mapping(input_mapping)
        self.frame_stack = None
        self.max_steps = max_steps
        self.observe_steps = observe_steps or (0.1 * replay_memory_size)
        self.current_observe_step = 0
        self.current_step = 0
        self.initial_epsilon = initial_epsilon
        self.final_epsilon = final_epsilon
        self.previous_epsilon = initial_epsilon
        self.epsilon_greedy_q_policy = EpsilonGreedyQPolicy(
            initial_epsilon=self.initial_epsilon,
            final_epsilon=self.final_epsilon,
            max_steps=self.max_steps
        )
        self.gamma = gamma
        self.current_action = None
        self.current_action_index = None
        self.current_action_type = None
        self.first_run = True
        self.mode = "OBSERVE"

        self.model_learning_rate = model_learning_rate
        self.model = self._initialize_model()

        if model_file_path is not None:
            self.load_model_weights(model_file_path, override_epsilon)

        self.model_loss = 0

        self.visual_debugger = VisualDebugger()

    def enter_train_mode(self):
        if self.previous_epsilon is not None:
            self.epsilon_greedy_q_policy.epsilon = self.previous_epsilon
            self.previous_epsilon = None

        self.mode = "TRAIN"

    def enter_run_mode(self):
        self.previous_epsilon = self.epsilon_greedy_q_policy.epsilon
        self.epsilon_greedy_q_policy.epsilon = 0.01
        self.mode = "RUN"

    def next_step(self):
        if self.mode == "TRAIN":
            self.current_step += 1
        elif self.mode == "OBSERVE":
            self.current_observe_step += 1

        if self.mode == "OBSERVE" and self.current_observe_step >= self.observe_steps:
            self.mode = "TRAIN"

    def build_frame_stack(self, game_frame):
        frame_stack = np.stack((
            game_frame,
            game_frame,
            game_frame,
            game_frame
        ), axis=2)

        self.frame_stack = frame_stack.reshape((1,) + frame_stack.shape)

    def update_frame_stack(self, game_frame_buffer):
        game_frames = [game_frame.frame for game_frame in game_frame_buffer.frames]
        frame_stack = np.stack(game_frames, axis=2)

        self.frame_stack = frame_stack.reshape((1,) + frame_stack.shape)

    def append_to_replay_memory(self, game_frame_buffer, reward, terminal=False):
        previous_frame_stack = self.frame_stack
        self.update_frame_stack(game_frame_buffer)

        observation = [
            previous_frame_stack,
            self.current_action_index,
            reward,
            self.frame_stack,
            terminal
        ]

        self.replay_memory.add(self.calculate_target_error(observation), observation)

    def calculate_target_error(self, observation):
        previous_target = self.model.predict(observation[0])[0][observation[1]]

        if observation[4]:
            target = observation[2]
        else:
            target = observation[2] + self.gamma * np.max(self.model.predict(observation[3]))

        return np.abs(target - previous_target)

    def pick_action(self, action_type=None):
        if action_type is None:
            self.compute_action_type()
        else:
            self.current_action_type = action_type

        qs = self.model.predict(self.frame_stack)

        if self.current_action_type == "RANDOM":
            self.current_action_index = random.randrange(self.action_count)
            self.maximum_future_rewards = None
        elif self.current_action_type == "PREDICTED":
            self.current_action_index = np.argmax(qs)
            self.maximum_future_rewards = qs

    def compute_action_type(self):
        use_random = self.epsilon_greedy_q_policy.use_random()
        self.current_action_type = "RANDOM" if use_random else "PREDICTED"

    def erode_epsilon(self, factor=1):
        if self.mode == "TRAIN":
            self.epsilon_greedy_q_policy.erode(factor=factor)

    def generate_mini_batch(self):
        if self.mode == "OBSERVE":
            return None

        return self.replay_memory.sample(self.batch_size)

    def train_on_mini_batch(self):
        mini_batch = self.generate_mini_batch()

        flashback_indices = random.sample(range(self.batch_size), 6)

        for i in range(0, len(mini_batch)):
            if i in flashback_indices:
                flashback_image = np.squeeze(mini_batch[i][1][3][:, :, :, 1])

                self.visual_debugger.store_image_data(
                    np.array(flashback_image * 255, dtype="uint8"),
                    flashback_image.shape,
                    f"flashback_{flashback_indices.index(i) + 1}"
                )

                del flashback_image

            previous_frame_stack = mini_batch[i][1][0]
            action_index = mini_batch[i][1][1]
            reward = mini_batch[i][1][2]
            frame_stack = mini_batch[i][1][3]
            terminal = mini_batch[i][1][4]

            target = self.model.predict(previous_frame_stack)
            previous_target = target[action_index]

            projected_future_rewards = self.model.predict(frame_stack)

            if terminal:
                target[action_index] = reward
            else:
                target[action_index] = reward + self.gamma * np.max(projected_future_rewards)

            error = np.abs(target[action_index] - previous_target)
            self.replay_memory.update(mini_batch[i][0], error)

            self.model.fit(previous_frame_stack, target, epochs=1, verbose=0)

    def generate_action(self):
        self.current_action = self.action_space.combinations[self.current_action_index]

    def get_action_for_index(self, action_index):
        return [action_input.upper() for action_input in self.action_input_mapping[self.action_space.combinations[action_index]]]

    def get_input_values(self):
        return self.action_input_mapping[self.current_action]

    def save_model_weights(self, file_path_prefix="datasets/model_", is_checkpoint=False):
        epsilon = self.epsilon_greedy_q_policy.epsilon

        if is_checkpoint:
            file_path = f"{file_path_prefix}_dqn_{self.current_step}_{epsilon}_.h5"
        else:
            file_path = f"{file_path_prefix}_dqn_{epsilon}_.h5"

        self.model.save_weights(file_path, overwrite=True)

    def load_model_weights(self, file_path, override_epsilon):
        self.model.load_weights(file_path)
        self.model.compile(loss="logcosh", optimizer=Adam(lr=self.model_learning_rate, clipvalue=10))

        *args, steps, epsilon, extension = file_path.split("_")
        self.current_step = int(steps)

        if override_epsilon:
            self.previous_epsilon = float(epsilon)
            self.epsilon_greedy_q_policy.epsilon = float(epsilon)

    def output_step_data(self):
        print(f"CURRENT MODE: {self.mode}")
        print(f"CURRENT STEP: {self.current_step}")

        if self.mode == "OBSERVE":
            print(f"CURRENT OBSERVE STEP: {self.current_observe_step}")
            print(f"OBSERVE STEPS: {self.observe_steps}")

        print(f"CURRENT EPSILON: {round(self.epsilon_greedy_q_policy.epsilon, 6)}")
        print(f"CURRENT RANDOM ACTION PROBABILITY: {round(self.epsilon_greedy_q_policy.epsilon * 100.0, 2)}%")
        print(f"LOSS: {self.model_loss}")

    def _initialize_model(self):
        input_layer = Input(shape=self.input_shape)

        tower_1 = Convolution2D(16, 1, 1, border_mode="same", activation="elu")(input_layer)
        tower_1 = Convolution2D(16, 3, 3, border_mode="same", activation="elu")(tower_1)

        tower_2 = Convolution2D(16, 1, 1, border_mode="same", activation="elu")(input_layer)
        tower_2 = Convolution2D(16, 3, 3, border_mode="same", activation="elu")(tower_2)
        tower_2 = Convolution2D(16, 3, 3, border_mode="same", activation="elu")(tower_2)

        tower_3 = MaxPooling2D((3, 3), strides=(1, 1), border_mode="same")(input_layer)
        tower_3 = Convolution2D(16, 1, 1, border_mode="same", activation="elu")(tower_3)

        merged_layer = merge([tower_1, tower_2, tower_3], mode="concat", concat_axis=1)

        output = AveragePooling2D((7, 7), strides=(8, 8))(merged_layer)
        output = Flatten()(output)
        output = Dense(self.action_count)(output)

        model = Model(input=input_layer, output=output)
        model.compile(rmsprop(lr=self.model_learning_rate, clipvalue=1), "mse")

        return model

    def _generate_action_space_combination_input_mapping(self, input_mapping):
        action_input_mapping = dict()

        for combination in self.action_space.combinations:
            combination_values = self.action_space.values_for_combination(combination)
            input_values = [input_mapping[combination_value] for combination_value in combination_values if combination_value is not None]

            action_input_mapping[combination] = list(itertools.chain.from_iterable(input_values))

        return action_input_mapping



import warnings
warnings.simplefilter("ignore")

import offshoot

import subprocess
import signal
import shlex
import time
import os, os.path
import atexit

from serpent.game_agent import GameAgent

from serpent.game_launchers import *

from serpent.window_controller import WindowController

from serpent.input_controller import InputController, InputControllers

from serpent.frame_grabber import FrameGrabber
from serpent.game_frame_limiter import GameFrameLimiter

from serpent.sprite import Sprite

from serpent.utilities import clear_terminal, is_windows

import skimage.io
import skimage.color

import numpy as np

from redis import StrictRedis

from serpent.config import config


class GameError(BaseException):
    pass


class Game(offshoot.Pluggable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.config = config.get(f"{self.__class__.__name__}Plugin")

        self.platform = kwargs.get("platform")

        default_input_controller_backend = InputControllers.NATIVE_WIN32 if is_windows() else InputControllers.PYAUTOGUI
        self.input_controller = kwargs.get("input_controller") or default_input_controller_backend

        self.window_id = None
        self.window_name = kwargs.get("window_name")
        self.window_geometry = None

        self.window_controller = WindowController()

        self.is_launched = False

        self.frame_grabber_process = None
        self.frame_transformation_pipeline_string = None

        self.game_frame_limiter = GameFrameLimiter(fps=self.config.get("fps", 30))

        self.api_class = None
        self.api_instance = None

        self.sprites = self._discover_sprites()

        self.redis_client = StrictRedis(**config["redis"])

        self.kwargs = kwargs

    @property
    @offshoot.forbidden
    def game_launcher(self):
        return self.game_launchers.get(self.platform)

    @property
    @offshoot.forbidden
    def game_launchers(self):
        return {
            "steam": SteamGameLauncher,
            "executable": ExecutableGameLauncher,
            "web_browser": WebBrowserGameLauncher
        }

    @property
    @offshoot.expected
    def screen_regions(self):
        raise NotImplementedError()

    @property
    @offshoot.expected
    def ocr_presets(self):
        raise NotImplementedError()

    @property
    @offshoot.forbidden
    def api(self):
        if self.api_instance is None:
            self.api_instance = self.api_class(game=self)
        else:
            return self.api_instance

    @property
    @offshoot.forbidden
    def is_focused(self):
        return self.window_controller.is_window_focused(self.window_id)

    @offshoot.forbidden
    def launch(self, dry_run=False):
        self.before_launch()

        if not dry_run:
            self.game_launcher().launch(**self.kwargs)

        self.after_launch()

    def before_launch(self):
        pass

    def after_launch(self):
        self.is_launched = True

        time.sleep(5)

        self.window_id = self.window_controller.locate_window(self.window_name)

        self.window_controller.move_window(self.window_id, 0, 0)
        self.window_controller.focus_window(self.window_id)

        self.window_geometry = self.extract_window_geometry()

        print(self.window_geometry)

    def play(self, game_agent_class_name="GameAgent", frame_handler=None, **kwargs):
        if not self.is_launched:
            raise GameError(f"Game '{self.__class__.__name__}' is not running...")

        game_agent_class = offshoot.discover("GameAgent", selection=game_agent_class_name).get(game_agent_class_name, GameAgent)

        if game_agent_class is None:
            raise GameError("The provided Game Agent class name does not map to an existing class...")

        game_agent = game_agent_class(
            game=self,
            input_controller=InputController(game=self, backend=self.input_controller),
            **kwargs
        )

        # Look if we need to auto-append PNG to frame transformation pipeline based on given frame_handler
        png_frame_handlers = ["RECORD"]

        if frame_handler in png_frame_handlers and self.frame_transformation_pipeline_string is not None:
            if not self.frame_transformation_pipeline_string.endswith("|PNG"):
                self.frame_transformation_pipeline_string += "|PNG"

        self.start_frame_grabber()
        self.redis_client.delete(config["frame_grabber"]["redis_key"])

        while self.redis_client.llen(config["frame_grabber"]["redis_key"]) == 0:
            time.sleep(0.1)

        self.window_controller.focus_window(self.window_id)

        frame_type = "FULL"

        pipeline_frame_handlers = [
            "COLLECT_FRAMES", 
            "COLLECT_FRAME_REGIONS", 
            "COLLECT_FRAMES_FOR_CONTEXT", 
            "RECORD"
        ]

        if frame_handler in pipeline_frame_handlers and self.frame_transformation_pipeline_string is not None:
            frame_type = "PIPELINE"

        # Override FPS Config?
        if frame_handler == "RECORD":
            self.game_frame_limiter = GameFrameLimiter(fps=10)

        try:
            while True:
                self.game_frame_limiter.start()

                game_frame = self.grab_latest_frame(frame_type=frame_type)

                try:
                    if self.is_focused:
                        game_agent.on_game_frame(game_frame, frame_handler=frame_handler, **kwargs)
                    else:
                        clear_terminal()
                        print("PAUSED\n")

                        game_agent.on_pause(frame_handler=frame_handler, **kwargs)

                        time.sleep(1)
                except Exception as e:
                    raise e
                    # print(e)
                    # time.sleep(0.1)

                self.game_frame_limiter.stop_and_delay()
        finally:
            self.stop_frame_grabber()

    @offshoot.forbidden
    def extract_window_geometry(self):
        if self.is_launched:
            return self.window_controller.get_window_geometry(self.window_id)

        return None

    @offshoot.forbidden
    def start_frame_grabber(self, pipeline_string=None):
        if not self.is_launched:
            raise GameError(f"Game '{self.__class__.__name__}' is not running...")

        if self.frame_grabber_process is not None:
            self.stop_frame_grabber()

        frame_grabber_command = f"serpent grab_frames {self.window_geometry['width']} {self.window_geometry['height']} {self.window_geometry['x_offset']} {self.window_geometry['y_offset']}"

        pipeline_string = pipeline_string or self.frame_transformation_pipeline_string

        if pipeline_string is not None:
            frame_grabber_command += f" {pipeline_string}"

        self.frame_grabber_process = subprocess.Popen(shlex.split(frame_grabber_command))

        signal.signal(signal.SIGINT, self._handle_signal)
        signal.signal(signal.SIGTERM, self._handle_signal)

        atexit.register(self._handle_signal, 15, None, False)

    @offshoot.forbidden
    def stop_frame_grabber(self):
        if self.frame_grabber_process is None:
            return None

        self.frame_grabber_process.kill()
        self.frame_grabber_process = None

        atexit.unregister(self._handle_signal)

    @offshoot.forbidden
    def grab_latest_frame(self, frame_type="FULL"):
        game_frame_buffer = FrameGrabber.get_frames([0], frame_type=frame_type)

        return game_frame_buffer.frames[0]

    def _discover_sprites(self):
        plugin_path = offshoot.config["file_paths"]["plugins"]
        sprites = dict()

        sprite_path = f"{plugin_path}/{self.__class__.__name__}Plugin/files/data/sprites"

        if os.path.isdir(sprite_path):
            files = os.scandir(sprite_path)

            for file in files:
                if file.name.endswith(".png"):
                    sprite_name = "_".join(file.name.split("/")[-1].split("_")[:-1]).replace(".png", "").upper()

                    sprite_image_data = skimage.io.imread(f"{sprite_path}/{file.name}")
                    sprite_image_data = sprite_image_data[..., np.newaxis]

                    if sprite_name not in sprites:
                        sprite = Sprite(sprite_name, image_data=sprite_image_data)
                        sprites[sprite_name] = sprite
                    else:
                        sprites[sprite_name].append_image_data(sprite_image_data)

        return sprites

    def _handle_signal(self, signum=15, frame=None, do_exit=True):
        if self.frame_grabber_process is not None:
            if self.frame_grabber_process.poll() is None:
                self.frame_grabber_process.send_signal(signum)

                if do_exit:
                    exit()



import offshoot

from serpent.config import config

import time
import uuid
import pickle
import h5py
import random

import subprocess
import signal
import shlex
import atexit

import os
import os.path

import serpent.cv

from serpent.utilities import clear_terminal, is_unix

from serpent.frame_grabber import FrameGrabber
from serpent.input_recorder import InputRecorder

from serpent.game_frame import GameFrame
from serpent.game_frame_buffer import GameFrameBuffer

from serpent.sprite_identifier import SpriteIdentifier

from serpent.visual_debugger.visual_debugger import VisualDebugger

import skimage.io
import skimage.transform
import skimage.util

from redis import StrictRedis

from datetime import datetime


class GameAgentError(BaseException):
    pass


class GameAgent(offshoot.Pluggable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.game = kwargs["game"]
        self.game.api

        self.config = config.get(f"{self.__class__.__name__}Plugin") or dict()

        self.redis_client = StrictRedis(**config["redis"])

        self.input_controller = kwargs["input_controller"]
        self.machine_learning_models = dict()

        self.frame_handlers = dict(
            NOOP=self.handle_noop,
            COLLECT_FRAMES=self.handle_collect_frames,
            COLLECT_FRAME_REGIONS=self.handle_collect_frame_regions,
            COLLECT_FRAMES_FOR_CONTEXT=self.handle_collect_frames_for_context,
            RECORD=self.handle_record
        )

        self.frame_handler_setups = dict(
            COLLECT_FRAMES=self.setup_collect_frames,
            COLLECT_FRAME_REGIONS=self.setup_collect_frame_regions,
            COLLECT_FRAMES_FOR_CONTEXT=self.setup_collect_frames_for_context,
            RECORD=self.setup_handle_record
        )

        self.frame_handler_pause_callbacks = dict(
            COLLECT_FRAMES=self.on_collect_frames_pause,
            COLLECT_FRAME_REGIONS=self.on_collect_frame_regions_pause,
            COLLECT_FRAMES_FOR_CONTEXT=self.on_collect_frames_for_context_pause,
            RECORD=self.on_record_pause
        )

        self.reward_functions = dict(
            TEST=self.reward_test   
        )

        self.frame_handler_setup_performed = False

        self.visual_debugger = VisualDebugger()

        self.game_frame_buffer = GameFrameBuffer(size=self.config.get("game_frame_buffer_size", 5))
        self.game_context = None

        self.sprite_identifier = SpriteIdentifier()
        self._register_sprites()

        self.flag = None

        self.uuid = str(uuid.uuid4())
        self.started_at = datetime.now()

        self.kwargs = kwargs

    @offshoot.forbidden
    def on_game_frame(self, game_frame, frame_handler=None, **kwargs):
        if not self.frame_handler_setup_performed:
            self._setup_frame_handler(frame_handler=frame_handler, **kwargs)

        frame_handler = self.frame_handlers.get(frame_handler or self.config.get("frame_handler", "NOOP"))

        frame_handler(game_frame, **kwargs)

        self.game_frame_buffer.add_game_frame(game_frame)

    @offshoot.forbidden
    def on_pause(self, frame_handler=None, **kwargs):
        on_frame_handler_pause = self.frame_handler_pause_callbacks.get(frame_handler or self.config.get("frame_handler"))

        if on_frame_handler_pause is not None:
            on_frame_handler_pause(**kwargs)

    @offshoot.forbidden
    def load_machine_learning_model(self, file_path):
        with open(file_path, "rb") as f:
            serialized_classifier = f.read()

        return pickle.loads(serialized_classifier)

    @offshoot.forbidden
    def update_game_frame(self, frame_type="FULL"):
        game_frame_buffer = FrameGrabber.get_frames([0], frame_type=frame_type)
        return game_frame_buffer.frames[0]

    def handle_noop(self, game_frame, **kwargs):
        time.sleep(1)

    def setup_collect_frames(self, **kwargs):
        self.game_frames = list()
        self.collected_frame_count = 0

    def setup_collect_frame_regions(self, **kwargs):
        self.game_frames = list()
        self.collected_frame_count = 0

    def setup_collect_frames_for_context(self, **kwargs):
        context = kwargs.get("context") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["context"]

        if not os.path.isdir(f"datasets/collect_frames_for_context/{context}"):
            os.mkdir(f"datasets/collect_frames_for_context/{context}")

        self.game_frames = list()
        self.collected_frame_count = 0

    def setup_handle_record(self, **kwargs):
        self.game_frame_buffers = list()
        self.input_recorder_process = None
        
        self.frame_offsets = list(range(0, (self.kwargs["frame_count"] * self.kwargs["frame_spacing"]) - 1, self.kwargs["frame_spacing"]))
        
        self._start_input_recorder()

        clear_terminal()
        print("Start playing the game! Focus out when you are done or want to save the collected data to that point.")

    def handle_collect_frames(self, game_frame, **kwargs):
        self.game_frames.append(game_frame)

        self.collected_frame_count += 1

        clear_terminal()
        print(f"Collected Frame #{self.collected_frame_count}")

        time.sleep(kwargs.get("interval") or self.config.get("collect_frames_interval") or 1)

    def handle_collect_frame_regions(self, game_frame, **kwargs):
        region = kwargs.get("region")

        self.game_frames.append(game_frame)

        self.collected_frame_count += 1

        clear_terminal()
        print(f"Collected Frame #{self.collected_frame_count} for Region: {region}")

        time.sleep(kwargs.get("interval") or self.config.get("collect_frames_interval") or 1)

    def handle_collect_frames_for_context(self, game_frame, **kwargs):
        context = kwargs.get("context") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["context"]
        interval = kwargs.get("interval") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["interval"]

        screen_region = kwargs.get("screen_region")

        if screen_region is not None:
            if screen_region not in self.game.screen_regions:
                raise GameAgentError("Invalid game screen region...")

            frame_region = serpent.cv.extract_region_from_image(
                game_frame.frame,
                self.game.screen_regions[screen_region]
            )

            game_frame = GameFrame(frame_region)

        self.game_frames.append(game_frame)

        self.collected_frame_count += 1

        clear_terminal()
        print(f"Collected Frame #{self.collected_frame_count} for Context: {context}")

        time.sleep(interval)

    def handle_record(self, game_frame, **kwargs):
        game_frame_buffer = FrameGrabber.get_frames(
            self.frame_offsets, 
            frame_type="PIPELINE"
        )

        self.game_frame_buffers.append(game_frame_buffer)
    
    def on_collect_frames_pause(self, **kwargs):
        for i, game_frame in enumerate(self.game_frames):
            print(f"Saving image {i + 1}/{len(self.game_frames)} to disk...")
            skimage.io.imsave(f"datasets/collect_frames/frame_{game_frame.timestamp}.png", game_frame.frame)

        self.game_frames = list()

    def on_collect_frame_regions_pause(self, **kwargs):
        region = kwargs["region"]
        region_path = f"datasets/collect_frames/{region}"

        for i, game_frame in enumerate(self.game_frames):
            frame_region = serpent.cv.extract_region_from_image(game_frame.frame, self.game.screen_regions.get(region))

            if not os.path.isdir(region_path):
                os.mkdir(region_path)

            print(f"Saving image {i + 1}/{len(self.game_frames)} to disk...")
            skimage.io.imsave(f"{region_path}/region_{game_frame.timestamp}.png", frame_region)

        self.game_frames = list()

    def on_collect_frames_for_context_pause(self, **kwargs):
        context = kwargs.get("context") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["context"]

        for i, game_frame in enumerate(self.game_frames):
            file_name = f"datasets/collect_frames_for_context/{context}/frame_{game_frame.timestamp}.png"

            print(f"Saving image {i + 1}/{len(self.game_frames)} to disk...")
            skimage.io.imsave(file_name, game_frame.frame)

        self.game_frames = list()

    def on_record_pause(self, **kwargs):
        InputRecorder.pause_input_recording()

        input_events = list()
        input_event_count = self.redis_client.llen(config["input_recorder"]["redis_key"])

        for i in range(input_event_count):
            input_events.append(pickle.loads(self.redis_client.lpop(config["input_recorder"]["redis_key"])))

        data = self._merge_frames_and_input_events(input_events)

        if not len(data):
            time.sleep(1)
            return None

        latest_game_frame_buffer = None

        active_keys = set()
        down_keys = dict()

        observations = dict()

        compute_reward = "reward_function" in self.config and self.config["reward_function"] in self.reward_functions
        reward_func = None 

        if compute_reward:
            reward_func = self.reward_functions[self.config["reward_function"]]

        for item in data:
            if isinstance(item, GameFrameBuffer):
                latest_game_frame_buffer = item

                reward_score = 0

                if compute_reward:
                    reward_score = reward_func(item.frames)
                
                timestamp = item.frames[-2].timestamp
                observations[timestamp] = [item, dict(), list(active_keys), list(), reward_score]
            elif item["type"] == "keyboard":
                key_name, key_event = item["name"].split("-")

                if key_event == "DOWN":
                    active_keys.add(key_name)

                    if latest_game_frame_buffer is not None:
                        timestamp = latest_game_frame_buffer.frames[-2].timestamp
                        observations[timestamp][1][key_name] = item["timestamp"]

                        down_keys[key_name] = timestamp

                elif key_event == "UP":
                    active_keys.remove(key_name)

                    if key_name in down_keys:
                        timestamp = down_keys[key_name]
                        
                        duration = item["timestamp"] - observations[timestamp][1][key_name]
                        observations[timestamp][1][key_name] = duration

                        del down_keys[key_name]
            elif item["type"] == "mouse":
                if latest_game_frame_buffer is not None:
                    timestamp = latest_game_frame_buffer.frames[-2].timestamp
                    observations[timestamp][3].append(item)

        print(f"Writing Recorded Input Data to 'datasets/input_recording.h5'... (0/{len(observations)})")

        with h5py.File("datasets/input_recording.h5", "a") as f:
            i = 0

            for timestamp, observation in observations.items():
                clear_terminal()
                print(f"Writing Recorded Input Data to 'datasets/input_recording.h5'... ({i + 1}/{len(observations)})")
                game_frame_buffer, keyboard_inputs, keyboard_inputs_active, mouse_inputs, reward_score = observation

                f.create_dataset(
                    f"{timestamp}-frames",
                    data=[game_frame.frame_bytes for game_frame in game_frame_buffer.frames]
                )

                f.create_dataset(
                    f"{timestamp}-keyboard-inputs",
                    data=[(key_name.encode("utf-8"), str(duration).encode("utf-8")) for key_name, duration in keyboard_inputs.items()]
                )

                f.create_dataset(
                    f"{timestamp}-keyboard-inputs-active",
                    data=[key_name.encode("utf-8") for key_name in keyboard_inputs_active]
                )

                filtered_mouse_inputs = list()
                mouse_move_index = None

                valid_game_window_x = range(
                    self.game.window_geometry["x_offset"],
                    self.game.window_geometry["x_offset"] + self.game.window_geometry["width"] + 1
                )

                valid_game_window_y = range(
                    self.game.window_geometry["y_offset"],
                    self.game.window_geometry["y_offset"] + self.game.window_geometry["height"] + 1
                )

                for mouse_input in mouse_inputs:
                    if mouse_input["x"] in valid_game_window_x and mouse_input["y"] in valid_game_window_y:
                        if mouse_input["name"] == "MOVE":
                            mouse_move_index = len(filtered_mouse_inputs)
                        
                        filtered_mouse_inputs.append(mouse_input)
                
                mouse_input_data = list()

                for i, mouse_input in enumerate(filtered_mouse_inputs):
                    if mouse_input["name"] == "MOVE" and i != mouse_move_index:
                        continue
                    
                    mouse_input_data.append((
                        mouse_input["name"].encode("utf-8"),
                        mouse_input["button"].encode("utf-8") if mouse_input["button"] else b"",
                        mouse_input["direction"].encode("utf-8") if mouse_input["direction"] else b"",
                        mouse_input["velocity"] or b"",
                        mouse_input["x"],
                        mouse_input["y"],
                        mouse_input["timestamp"]
                    ))

                f.create_dataset(
                    f"{timestamp}-mouse-inputs",
                    data=mouse_input_data
                )

                f.create_dataset(
                    f"{timestamp}-reward",
                    data=reward_score
                )

                i += 1

            self.game_frame_buffers = list()

            clear_terminal()
            print(f"Writing Frame/Input Data to 'datasets/input_recording.h5'... DONE")

        time.sleep(1)

    def reward_test(self, frames, **kwargs):
        return random.choice(range(0, 10))

    def _setup_frame_handler(self, frame_handler=None, **kwargs):
        frame_handler = frame_handler or self.config.get("frame_handler", "NOOP")

        if frame_handler in self.frame_handler_setups:
            self.frame_handler_setups[frame_handler](**kwargs)

        self.frame_handler_setup_performed = True

    def _register_sprites(self):
        for sprite_name, sprite in self.game.sprites.items():
            self.sprite_identifier.register(sprite)
    
    @offshoot.forbidden
    def _start_input_recorder(self):
        if self.input_recorder_process is not None:
            self._stop_input_recorder()

        input_recorder_command = "serpent record_inputs"

        self.input_recorder_process = subprocess.Popen(shlex.split(input_recorder_command))

        signal.signal(signal.SIGINT, self._handle_signal)
        signal.signal(signal.SIGTERM, self._handle_signal)

        atexit.register(self._handle_signal, 15, None, False)

    @offshoot.forbidden
    def _stop_input_recorder(self):
        if self.input_recorder_process is None:
            return None

        self.input_recorder_process.kill()
        self.input_recorder_process = None

        atexit.unregister(self._handle_signal)

    @offshoot.forbidden
    def _handle_signal(self, signum=15, frame=None, do_exit=True):
        if self.input_recorder_process is not None:
            if self.input_recorder_process.poll() is None:
                self.input_recorder_process.send_signal(signum)

                if do_exit:
                    exit()
    
    @offshoot.forbidden
    def _merge_frames_and_input_events(self, input_events):
        game_frame_buffer_index = 0
        input_event_index = 0

        merged = list()

        while True:
            game_frame_buffer = None
            input_event = None

            if game_frame_buffer_index > (len(self.game_frame_buffers) - 1) and input_event_index > (len(input_events) - 1):
                break
            else:
                if game_frame_buffer_index <= (len(self.game_frame_buffers) - 1):
                    game_frame_buffer = self.game_frame_buffers[game_frame_buffer_index]
                
                if input_event_index <= (len(input_events) - 1):
                    input_event = input_events[input_event_index]

            if game_frame_buffer is None:
                item = input_event
                input_event_index += 1
            elif input_event is None:
                item = game_frame_buffer
                game_frame_buffer_index += 1
            else:
                game_frame_buffer_timestamp = game_frame_buffer.frames[-2].timestamp
                input_event_timestamp = input_event["timestamp"]

                if game_frame_buffer_timestamp < input_event_timestamp:
                    item = game_frame_buffer
                    game_frame_buffer_index += 1
                else:
                    item = input_event
                    input_event_index += 1
                
            merged.append(item)

        return merged



import itertools


class KeyboardMouseActionSpace:

    def __init__(self, **kwargs):
        self.input_label_index_mapping = dict()
        self.input_discrete_spaces = list()

        for i, (input_label, inputs) in enumerate(kwargs.items()):
            self.input_label_index_mapping[input_label] = i
            self.input_discrete_spaces.append([0, len(inputs)])

        self.labeled_inputs = kwargs

        self.combination_values = None

    @property
    def combinations(self):
        if self.combination_values is None:
            input_groups = [inputs for inputs in self.labeled_inputs.values()]

            self.combination_values = list(
                itertools.product(*[list(map(lambda t: t[0], enumerate(input_group))) for input_group in input_groups])
            )

        return self.combination_values

    def values_for_combination(self, combination):
        values = list()

        for index, value_index in enumerate(combination):
            values.append((self.labeled_inputs[list(self.labeled_inputs.keys())[index]])[value_index])

        return values



from serpent.utilities import is_linux, is_macos, is_windows


class WindowControllerError(BaseException):
    pass


class WindowController:

    def __init__(self):
        self.adapter = self._load_adapter()()

    def locate_window(self, name):
        return self.adapter.locate_window(name)

    def move_window(self, window_id, x, y):
        self.adapter.move_window(window_id, x, y)

    def resize_window(self, window_id, width, height):
        self.adapter.resize_window(window_id, width, height)

    def focus_window(self, window_id):
        self.adapter.focus_window(window_id)

    def is_window_focused(self, window_id):
        return self.adapter.is_window_focused(window_id)

    def get_focused_window_name(self):
        return self.adapter.get_focused_window_name()

    def get_window_geometry(self, window_id):
        return self.adapter.get_window_geometry(window_id)

    def _load_adapter(self):
        if is_linux():
            from serpent.window_controllers.linux_window_controller import LinuxWindowController
            return LinuxWindowController
        elif is_macos():
            from serpent.window_controllers.darwin_window_controller import DarwinWindowController
            return DarwinWindowController
        elif is_windows():
            from serpent.window_controllers.win32_window_controller import Win32WindowController
            return Win32WindowController



from serpent.game_launcher import GameLauncher, GameLauncherException

from serpent.utilities import is_linux, is_macos, is_windows

import enum
import webbrowser


class WebBrowser(enum.Enum):
    DEFAULT = 0
    CHROME = 1
    CHROMIUM = 2
    FIREFOX = 3
    OPERA = 4
    SAFARI = 5


class WebBrowserGameLauncher(GameLauncher):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def web_browsers(self):
        return {
            WebBrowser.DEFAULT.name: None,
            WebBrowser.CHROME.name: "chrome",
            WebBrowser.CHROMIUM.name: "chromium",
            WebBrowser.FIREFOX.name: "firefox",
            WebBrowser.OPERA.name: "opera",
            WebBrowser.SAFARI.name: "safari"
        }

    def launch(self, **kwargs):
        url = kwargs.get("url")
        browser = kwargs.get("browser") or WebBrowser.DEFAULT

        if url is None:
            raise GameLauncherException("An 'url' kwarg is required...")

        if is_linux():
            webbrowser.get(self.web_browsers.get(browser.name)).open_new(url)
        elif is_macos():
            webbrowser.get(self.web_browsers.get(browser.name)).open_new(url)
        elif is_windows():
            webbrowser.get(self.web_browsers.get(browser.name)).open_new(url)



import asyncio

from autobahn.asyncio.wamp import ApplicationSession, ApplicationRunner
from autobahn.wamp import auth

from elasticsearch import Elasticsearch

from serpent.config import config

import uuid


class AnalyticsElasticsearchComponent:
    @classmethod
    def run(cls):
        print(f"Starting {cls.__name__}...")

        url = "ws://%s:%s" % (config["analytics"]["host"], config["analytics"]["port"])

        runner = ApplicationRunner(url=url, realm=config["analytics"]["realm"])
        runner.run(AnalyticsElasticsearchWAMPComponent)


class AnalyticsElasticsearchWAMPComponent(ApplicationSession):
    def __init__(self, c=None):
        super().__init__(c)

        self.es_client = Elasticsearch(hosts=config["elasticsearch"]["hosts"])
        self.event_buffer = list()

    def onConnect(self):
        self.join(config["analytics"]["realm"], ["wampcra"], config["analytics"]["auth"]["username"])

    def onDisconnect(self):
        print("Disconnected from Crossbar!")

    def onChallenge(self, challenge):
        secret = config["analytics"]["auth"]["password"]
        signature = auth.compute_wcs(secret.encode('utf8'), challenge.extra['challenge'].encode('utf8'))

        return signature.decode('ascii')

    async def onJoin(self, details):

        async def on_event_message(event):
            if not isinstance(event, dict):
                return None

            event["uuid"] = str(uuid.uuid4())

            self.es_client.index(
                index=f"serpent:{config['analytics']['topic']}:events".lower(),
                doc_type="event",
                body=event,
                id=event["uuid"]
            )

        self.subscribe(
            on_event_message,
            config["analytics"]["topic"]
        )



import enum

from serpent.utilities import is_linux, is_macos, is_windows

from sneakysnek.keyboard_keys import KeyboardKey
from sneakysnek.mouse_buttons import MouseButton


# US layout - What's to be done for other keyboard layouts?
character_keyboard_key_mapping = {
    "`": [KeyboardKey.KEY_GRAVE],
    "~": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_GRAVE],
    "1": [KeyboardKey.KEY_1],
    "!": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_1],
    "2": [KeyboardKey.KEY_2],
    "@": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_2],
    "3": [KeyboardKey.KEY_3],
    "#": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_3],
    "4": [KeyboardKey.KEY_4],
    "$": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_4],
    "5": [KeyboardKey.KEY_5],
    "%": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_5],
    "6": [KeyboardKey.KEY_6],
    "^": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_6],
    "7": [KeyboardKey.KEY_7],
    "&": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_7],
    "8": [KeyboardKey.KEY_8],
    "*": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_8],
    "9": [KeyboardKey.KEY_9],
    "(": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_9],
    "0": [KeyboardKey.KEY_0],
    ")": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_0],
    "-": [KeyboardKey.KEY_MINUS],
    "_": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_MINUS],
    "=": [KeyboardKey.KEY_EQUALS],
    "+": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_EQUALS],
    "q": [KeyboardKey.KEY_Q],
    "Q": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_Q],
    "w": [KeyboardKey.KEY_W],
    "W": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_W],
    "e": [KeyboardKey.KEY_E],
    "E": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_E],
    "r": [KeyboardKey.KEY_R],
    "R": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_R],
    "t": [KeyboardKey.KEY_T],
    "T": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_T],
    "y": [KeyboardKey.KEY_Y],
    "Y": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_Y],
    "u": [KeyboardKey.KEY_U],
    "U": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_U],
    "i": [KeyboardKey.KEY_I],
    "I": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_I],
    "o": [KeyboardKey.KEY_O],
    "O": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_O],
    "p": [KeyboardKey.KEY_P],
    "P": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_P],
    "[": [KeyboardKey.KEY_LEFT_BRACKET],
    "{": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_LEFT_BRACKET],
    "]": [KeyboardKey.KEY_RIGHT_BRACKET],
    "}": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_RIGHT_BRACKET],
    "\n": [KeyboardKey.KEY_RETURN],
    "a": [KeyboardKey.KEY_A],
    "A": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_A],
    "s": [KeyboardKey.KEY_S],
    "S": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_S],
    "d": [KeyboardKey.KEY_D],
    "D": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_D],
    "f": [KeyboardKey.KEY_F],
    "F": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_F],
    "g": [KeyboardKey.KEY_G],
    "G": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_G],
    "h": [KeyboardKey.KEY_H],
    "H": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_H],
    "j": [KeyboardKey.KEY_J],
    "J": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_J],
    "k": [KeyboardKey.KEY_K],
    "K": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_K],
    "l": [KeyboardKey.KEY_L],
    "L": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_L],
    ";": [KeyboardKey.KEY_SEMICOLON],
    ":": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_SEMICOLON],
    "'": [KeyboardKey.KEY_APOSTROPHE],
    '"': [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_APOSTROPHE],
    "z": [KeyboardKey.KEY_Z],
    "Z": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_Z],
    "x": [KeyboardKey.KEY_X],
    "X": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_X],
    "c": [KeyboardKey.KEY_C],
    "C": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_C],
    "v": [KeyboardKey.KEY_V],
    "V": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_V],
    "b": [KeyboardKey.KEY_B],
    "B": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_B],
    "n": [KeyboardKey.KEY_N],
    "N": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_N],
    "m": [KeyboardKey.KEY_M],
    "M": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_M],
    ",": [KeyboardKey.KEY_COMMA],
    "<": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_COMMA],
    ".": [KeyboardKey.KEY_PERIOD],
    ">": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_PERIOD],
    "/": [KeyboardKey.KEY_SLASH],
    "?": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_SLASH],
    "\\": [KeyboardKey.KEY_BACKSLASH],
    "|": [KeyboardKey.KEY_LEFT_SHIFT, KeyboardKey.KEY_BACKSLASH],
    " ": [KeyboardKey.KEY_SPACE]
}


class InputControllers(enum.Enum):
    PYAUTOGUI = 0
    NATIVE_WIN32 = 1


class InputControllerError(BaseException):
    pass


class InputController:

    def __init__(self, backend=InputControllers.PYAUTOGUI, game=None, **kwargs):
        self.game = game
        self.backend = self._initialize_backend(backend, **kwargs)

    @property
    def game_is_focused(self):
        return self.game.is_focused

    # Keyboard Actions
    def handle_keys(self, key_collection, **kwargs):
        self._is_game_launched_check()
        self.backend.handle_keys(key_collection, **kwargs)

    def tap_keys(self, keys, duration=0.05, **kwargs):
        self._is_game_launched_check()
        self.backend.tap_keys(keys, duration=duration, **kwargs)

    def tap_key(self, key, duration=0.05, **kwargs):
        self._is_game_launched_check()
        self.backend.tap_key(key, duration=duration, **kwargs)

    def press_keys(self, keys, **kwargs):
        self._is_game_launched_check()
        self.backend.press_keys(keys, **kwargs)

    def press_key(self, key, **kwargs):
        self._is_game_launched_check()
        self.backend.press_key(key, **kwargs)

    def release_keys(self, keys, **kwargs):
        self._is_game_launched_check()
        self.backend.release_keys(keys, **kwargs)

    def release_key(self, key, **kwargs):
        self._is_game_launched_check()
        self.backend.release_key(key, **kwargs)

    def type_string(self, string, duration=0.05, **kwargs):
        self._is_game_launched_check()
        self.backend.type_string(string, duration=duration, **kwargs)

    # Mouse Actions
    def move(self, x=None, y=None, duration=0.25, absolute=True, **kwargs):
        self._is_game_launched_check()
        self.backend.move(x=x, y=y, duration=duration, absolute=absolute, **kwargs)

    def click_down(self, button=MouseButton.LEFT, **kwargs):
        self._is_game_launched_check()
        self.backend.click_down(button=button, **kwargs)

    def click_up(self, button=MouseButton.LEFT, **kwargs):
        self._is_game_launched_check()
        self.backend.click_up(button=button, **kwargs)

    def click(self, button=MouseButton.LEFT, duration=0.25, **kwargs):
        self._is_game_launched_check()
        self.backend.click(button=button, duration=duration, **kwargs)

    def click_screen_region(self, button=MouseButton.LEFT, screen_region=None, **kwargs):
        self._is_game_launched_check()
        self.backend.click_screen_region(button=button, screen_region=screen_region, **kwargs)

    def click_sprite(self, button=MouseButton.LEFT, sprite=None, game_frame=None, **kwargs):
        self._is_game_launched_check()
        self.backend.click_sprite(button=button, sprite=sprite, game_frame=game_frame, **kwargs)

    # Requires the Serpent OCR module
    def click_string(self, query_string, button=MouseButton.LEFT, game_frame=None, fuzziness=2, ocr_preset=None, **kwargs):
        self._is_game_launched_check()
        self.backend.click_string(query_string, button=button, game_frame=game_frame, fuzziness=fuzziness, ocr_preset=ocr_preset, **kwargs)

    def drag(self, button=MouseButton.LEFT, x0=None, y0=None, x1=None, y1=None, duration=1, **kwargs):
        self._is_game_launched_check()
        self.backend.drag(button=button, x0=x0, y0=y0, x1=x1, y1=y1, duration=duration, **kwargs)

    def drag_screen_region_to_screen_region(self, button=MouseButton.LEFT, start_screen_region=None, end_screen_region=None, duration=1, **kwargs):
        self._is_game_launched_check()
        self.backend.drag_screen_region_to_screen_region(button=button, start_screen_region=start_screen_region, end_screen_region=end_screen_region, duration=duration, **kwargs)

    def scroll(self, x=None, y=None, clicks=1, direction="DOWN", **kwargs):
        self._is_game_launched_check()
        self.backend.scroll(x=x, y=y, clicks=clicks, direction=direction, **kwargs)

    def _initialize_backend(self, backend, **kwargs):
        if backend == InputControllers.PYAUTOGUI:
            from serpent.input_controllers.pyautogui_input_controller import PyAutoGUIInputController
            return PyAutoGUIInputController(game=self.game, **kwargs)
        elif backend == InputControllers.NATIVE_WIN32:
            from serpent.input_controllers.native_win32_input_controller import NativeWin32InputController
            return NativeWin32InputController(game=self.game, **kwargs)
        else:
            raise InputControllerError("The specified backend is invalid!")

    def _is_game_launched_check(self):
        if not self.game.is_launched:
            raise InputControllerError("InputController cannot be used while the game is not running!")

    def _extract_screen_region_coordinates(self, screen_region):
        screen_region_coordinates = self.game.screen_regions.get(screen_region)

        x = (screen_region_coordinates[1] + screen_region_coordinates[3]) // 2
        x += self.game.window_geometry["x_offset"]

        y = (screen_region_coordinates[0] + screen_region_coordinates[2]) // 2
        y += self.game.window_geometry["y_offset"]

        return x, y



import skimage.color
import skimage.measure
import skimage.transform
import skimage.filters
import skimage.morphology

import numpy as np


class GameFrameError(BaseException):
    pass


class GameFrame:

    def __init__(self, frame_data, frame_variants=None, timestamp=None, **kwargs):
        if isinstance(frame_data, bytes):
            self.frame_bytes = frame_data
            self.frame_array = None
        elif isinstance(frame_data, np.ndarray):
            self.frame_bytes = None
            self.frame_array = frame_data

        self.frame_variants = frame_variants or dict()

        self.timestamp = timestamp

        self.offset_x = kwargs.get("offset_x") or 0
        self.offset_y = kwargs.get("offset_y") or 0

        self.resize_order = kwargs.get("resize_order") or 1

    @property
    def frame(self):
        return self.frame_array if self.frame_array is not None else self.frame_bytes

    @property
    def half_resolution_frame(self):
        """ A quarter-sized version of the frame (half-width, half-height)"""

        if "half" not in self.frame_variants:
            self.frame_variants["half"] = self._to_half_resolution()

        return self.frame_variants["half"]

    @property
    def quarter_resolution_frame(self):
        """ A sixteenth-sized version of the frame (quarter-width, quarter-height)"""

        if "quarter" not in self.frame_variants:
            self.frame_variants["quarter"] = self._to_quarter_resolution()

        return self.frame_variants["quarter"]

    @property
    def eighth_resolution_frame(self):
        """ A 1/32-sized version of the frame (eighth-width, eighth-height)"""

        if "eighth" not in self.frame_variants:
            self.frame_variants["eighth"] = self._to_eighth_resolution()

        return self.frame_variants["eighth"]

    @property
    def eighth_resolution_grayscale_frame(self):
        """ A 1/32-sized, grayscale version of the frame (eighth-width, eighth-height)"""

        if "eighth_grayscale" not in self.frame_variants:
            self.frame_variants["eighth_grayscale"] = self._to_eighth_grayscale_resolution()

        return self.frame_variants["eighth_grayscale"]

    @property
    def grayscale_frame(self):
        """ A full-size grayscale version of the frame"""

        if "grayscale" not in self.frame_variants:
            self.frame_variants["grayscale"] = self._to_grayscale()

        return self.frame_variants["grayscale"]

    @property
    def ssim_frame(self):
        """ A 100x100 grayscale frame to be used for SSIM"""

        if "ssim" not in self.frame_variants:
            self.frame_variants["ssim"] = self._to_ssim()

        return self.frame_variants["ssim"]

    def compare_ssim(self, previous_game_frame):
        return skimage.measure.compare_ssim(previous_game_frame.ssim_frame, self.ssim_frame)

    def difference(self, previous_game_frame):
        current = skimage.filters.gaussian(self.grayscale_frame, 8)
        previous = skimage.filters.gaussian(previous_game_frame.grayscale_frame, 8)

        return current - previous

    # TODO: Refactor Fraction of Resolution Frames...
    def _to_half_resolution(self):
        shape = (
            self.frame_array.shape[0] // 2,
            self.frame_array.shape[1] // 2
        )

        return np.array(skimage.transform.resize(self.frame_array, shape, mode="reflect", order=self.resize_order) * 255, dtype="uint8")

    def _to_quarter_resolution(self):
        shape = (
            self.frame_array.shape[0] // 4,
            self.frame_array.shape[1] // 4
        )

        return np.array(skimage.transform.resize(self.frame_array, shape, mode="reflect", order=self.resize_order) * 255, dtype="uint8")

    def _to_eighth_resolution(self):
        shape = (
            self.frame_array.shape[0] // 8,
            self.frame_array.shape[1] // 8
        )

        return np.array(skimage.transform.resize(self.frame_array, shape, mode="reflect", order=self.resize_order) * 255, dtype="uint8")

    def _to_eighth_grayscale_resolution(self):
        shape = (
            self.frame_array.shape[0] // 8,
            self.frame_array.shape[1] // 8
        )

        return np.array(skimage.transform.resize(self.grayscale_frame, shape, mode="reflect", order=self.resize_order) * 255, dtype="uint8")

    def _to_grayscale(self):
        return np.array(skimage.color.rgb2gray(self.frame_array) * 255, dtype="uint8")

    def _to_ssim(self):
        grayscale = self.grayscale_frame
        return skimage.transform.resize(grayscale, (100, 100), mode="reflect", order=0)



from serpent.visual_debugger.visual_debugger import VisualDebugger

from serpent.utilities import SerpentError

from serpent.config import config

try:
    from kivy.app import App

    from kivy.core.window import Window
    from kivy.core.image import Image as CoreImage

    from kivy.uix.widget import Widget
    from kivy.uix.image import Image
    from kivy.uix.label import Label

    from kivy.uix.floatlayout import FloatLayout
    from kivy.uix.gridlayout import GridLayout
    from kivy.uix.boxlayout import BoxLayout

    from kivy.clock import Clock
except ImportError:
    raise SerpentError("Setup has not been been performed for the GUI module. Please run 'serpent setup gui'")

from PIL import Image as PILImage

import io


class VisualDebuggerApp(App):
    def __init__(self, buckets=None):
        super().__init__()

        self.visual_debugger = VisualDebugger(buckets=buckets)
        self.canvas = None

    def build(self):
        self.canvas = VisualDebuggerCanvas(buckets=self.visual_debugger.available_buckets)

        Clock.schedule_interval(self.update_image_data, 0.01)

        return self.canvas

    def update_image_data(self, *args):
        response = self.visual_debugger.retrieve_image_data()

        if response is not None:
            bucket, image_data = response
            self.canvas.update(bucket, image_data)


class VisualDebuggerCanvas(Widget):

    def __init__(self, buckets=None):
        super().__init__()

        self.images = dict()

        self.root = FloatLayout(size=(Window.width, Window.height))
        self.grid = GridLayout(cols=8)

        self.add_widget(self.root)
        self.root.add_widget(self.grid)

        for i, bucket in enumerate(buckets):
            layout = BoxLayout(orientation="vertical")

            image = VisualDebuggerImage(
                allow_stretch=True
            )

            image.bind(texture=image.update_texture_filters)

            self.images[bucket] = image

            layout.add_widget(image)

            label = Label(
                text=bucket,
                color=(1, 1, 1, 1),
                size_hint=(1, 0.1)
            )

            layout.add_widget(label)

            self.grid.add_widget(layout)

        Window.bind(on_resize=self.on_window_resize)
        Window.clearcolor = (0.136, 0.191, 0.25, 1)

    def update(self, bucket, image_data):
        image = PILImage.fromarray(image_data).convert("RGB")
        image_file = io.BytesIO()

        image.save(image_file, "png")
        image_file.seek(0)

        core_image = CoreImage(image_file, ext="png")
        self.images[bucket].texture = core_image.texture

    def on_window_resize(self, window, width, height):
        self.root.size = (width, height)


class VisualDebuggerImage(Image):

    def update_texture_filters(self, image, texture):
        if not texture:
            return

        texture.min_filter = 'nearest'
        texture.mag_filter = 'nearest'



import pickle
import time

from redis import StrictRedis

from serpent.config import config

from sneakysnek.recorder import Recorder
from sneakysnek.keyboard_event import KeyboardEvent, KeyboardEvents
from sneakysnek.mouse_event import MouseEvent


redis_client = StrictRedis(**config["redis"])


class InputRecorder:

    redis_key = config["input_recorder"]["redis_key"]
    redis_key_pause = f"{redis_key}:PAUSED"
    redis_key_stop = f"{redis_key}:STOPPED"

    def __init__(self):
        self.active_keys = set()
        self.recorder = None

        self.redis_key = self.__class__.redis_key
        self.redis_key_pause = self.__class__.redis_key_pause
        self.redis_key_stop = self.__class__.redis_key_stop
        
        redis_client.delete(self.redis_key)
        redis_client.delete(self.redis_key_pause)
        redis_client.delete(self.redis_key_stop)

    def start(self):
        self.recorder = Recorder.record(self._on_input_event)

        while True:
            time.sleep(1)

    def stop(self):
        self.recorder.stop()

    def _on_input_event(self, event):
        if redis_client.get(self.redis_key_pause) == b"1":
            return None

        if redis_client.get(self.redis_key_stop) == b"1":
            self.stop()
            return None

        if isinstance(event, KeyboardEvent):
            if event.event == KeyboardEvents.DOWN:
                if event.keyboard_key.name in self.active_keys:
                    return None

                self.active_keys.add(event.keyboard_key.name)
            elif event.event == KeyboardEvents.UP:
                if event.keyboard_key.name in self.active_keys:
                    self.active_keys.remove(event.keyboard_key.name)

            event = {
                "type": "keyboard",
                "name": f"{event.keyboard_key.name}-{event.event.value.upper()}", 
                "timestamp": event.timestamp
            }

            event = pickle.dumps(event)
            redis_client.rpush(config["input_recorder"]["redis_key"], event)
        elif isinstance(event, MouseEvent):
            event = {
                "type": "mouse",
                "name": event.event.name,
                "button": event.button.name if event.button else None,
                "direction": event.direction if event.direction else None,
                "velocity": event.velocity if event.velocity else None,
                "x": event.x,
                "y": event.y,
                "timestamp": event.timestamp
            }

            event = pickle.dumps(event)
            redis_client.rpush(config["input_recorder"]["redis_key"], event)

    @classmethod
    def pause_input_recording(cls):
        redis_client.set(cls.redis_key_pause, 1)

    @classmethod
    def resume_input_recording(cls):
        redis_client.set(cls.redis_key_pause, 0)

    @classmethod
    def stop_input_recording(cls):
        redis_client.set(cls.redis_key_stop, 1)



import skimage.measure


class SpriteIdentifier:
    def __init__(self):
        self.sprites = dict()

    def identify(self, sprite, mode="SIGNATURE_COLORS", score_threshold=75, debug=False):
        if mode == "SIGNATURE_COLORS":
            return self.identify_by_signature_colors(sprite, score_threshold=score_threshold, debug=debug)
        elif mode == "CONSTELLATION_OF_PIXELS":
            return self.identify_by_constellation_of_pixels(sprite, score_threshold=score_threshold, debug=debug)
        elif mode == "SSIM":
            return self.identify_by_ssim(sprite, score_threshold=score_threshold, debug=debug)

    def identify_by_signature_colors(self, query_sprite, score_threshold=0, debug=False):
        top_sprite_score = 0
        top_sprite_match = None

        for sprite_name, sprite in self.sprites.items():
            for sprite_signature_colors in sprite.signature_colors:
                for query_sprite_signature_colors in query_sprite.signature_colors:
                    signature_color_score = int((len(query_sprite_signature_colors & sprite_signature_colors) / len(sprite_signature_colors)) * 100)

                    if debug:
                        print(sprite_name, signature_color_score)

                    if signature_color_score > top_sprite_score:
                        top_sprite_score = signature_color_score
                        top_sprite_match = sprite_name

        return top_sprite_match if top_sprite_score >= score_threshold else "UNKNOWN"

    def identify_by_constellation_of_pixels(self, query_sprite, score_threshold=0, debug=False):
        top_sprite_score = 0
        top_sprite_match = None

        for sprite_name, sprite in self.sprites.items():
            if sprite.image_shape != query_sprite.image_shape:
                continue

            for constellation_of_pixels in sprite.constellation_of_pixels:
                for i in range(query_sprite.image_data.shape[3]):
                    query_sprite_image = query_sprite.image_data[..., i]

                    constellation_of_pixels_score = 0

                    for pixel_coordinates, pixel_color in constellation_of_pixels.items():
                        if tuple(query_sprite_image[pixel_coordinates[0], pixel_coordinates[1], :][:3]) == pixel_color:
                            constellation_of_pixels_score += 1

                    constellation_of_pixels_score = int((constellation_of_pixels_score / len(constellation_of_pixels)) * 100)

                    if debug:
                        print(sprite_name, constellation_of_pixels_score)

                    if constellation_of_pixels_score > top_sprite_score:
                        top_sprite_score = constellation_of_pixels_score
                        top_sprite_match = sprite_name

        return top_sprite_match if top_sprite_score >= score_threshold else "UNKNOWN"

    def identify_by_ssim(self, query_sprite, score_threshold=0, debug=False):
        top_sprite_score = 0
        top_sprite_match = None

        for sprite_name, sprite in self.sprites.items():
            if sprite.image_data[..., 0].shape[:2] == query_sprite.image_data[..., 0].shape[:2]:
                for i in range(sprite.image_data.shape[3]):
                    for ii in range(query_sprite.image_data.shape[3]):
                        sprite_image = sprite.image_data[..., :3, i]
                        query_sprite_image = query_sprite.image_data[..., :3, ii]

                        ssim_score = int(skimage.measure.compare_ssim(query_sprite_image, sprite_image, multichannel=True) * 100)

                        if debug:
                            print(sprite_name, ssim_score)

                        if ssim_score > top_sprite_score:
                            top_sprite_score = ssim_score
                            top_sprite_match = sprite_name
            else:
                if debug:
                    print(f"The shape of '{sprite_name}' does not match the query sprite's shape. Skipping!")

        return top_sprite_match if top_sprite_score >= score_threshold else "UNKNOWN"

    def register(self, sprite):
        self.sprites[sprite.name] = sprite



#!/usr/bin/env python
import os
import sys
import shutil
import subprocess
import shlex
import time

import offshoot

from serpent.utilities import clear_terminal, display_serpent_logo, is_linux, is_macos, is_windows, is_unix

from serpent.window_controller import WindowController

# Add the current working directory to sys.path to discover user plugins!
sys.path.insert(0, os.getcwd())

VERSION = "2018.1.0"

valid_commands = [
    "setup",
    "update",
    "modules",
    "grab_frames",
    "launch",
    "play",
    "record",
    "generate",
    "activate",
    "deactivate",
    "plugins",
    "train",
    "capture",
    "visual_debugger",
    "window_name",
    "record_inputs"
]


def execute():
    if len(sys.argv) == 1:
        executable_help()
    elif len(sys.argv) > 1:
        if sys.argv[1] == "-h" or sys.argv[1] == "--help":
            executable_help()
        else:
            command = sys.argv[1]

            if command not in valid_commands:
                raise Exception("'%s' is not a valid Serpent command." % command)

            command_function_mapping[command](*sys.argv[2:])


def executable_help():
    print(f"\nSerpent.AI v{VERSION}")
    print("Available Commands:\n")

    for command, description in command_description_mapping.items():
        print(f"{command.rjust(16)}: {description}")

    print("")


def setup(module=None):
    clear_terminal()
    display_serpent_logo()
    print("")

    if module is None:
        setup_base()
    elif module == "ocr":
        setup_ocr()
    elif module == "gui":
        setup_gui()
    elif module == "ml":
        setup_ml()
    else:
        print(f"Invalid Setup Module: {module}")

def setup_base():
    # Copy Config Templates
    print("Creating Configuration Files...")

    first_run = True

    for root, directories, files in os.walk(os.getcwd()):
        for file in files:
            if file == "offshoot.yml":
                first_run = False
                break

        if not first_run:
            break

    if not first_run:
        confirm = input("It appears that the setup process had already been performed. Are you sure you want to proceed? Some important files will be overwritten! (One of: 'YES', 'NO'):\n")

        if confirm not in ["YES", "NO"]:
            confirm = "NO"

        if confirm == "NO":
            sys.exit()

    shutil.rmtree(os.path.join(os.getcwd(), "config"), ignore_errors=True)

    shutil.copytree(
        os.path.join(os.path.dirname(__file__), "config"),
        os.path.join(os.getcwd(), "config")
    )

    # Copy Offshoot Files
    shutil.copy(
        os.path.join(os.path.dirname(__file__), "offshoot.yml"),
        os.path.join(os.getcwd(), "offshoot.yml")
    )

    shutil.copy(
        os.path.join(os.path.dirname(__file__), "offshoot.manifest.json"),
        os.path.join(os.getcwd(), "offshoot.manifest.json")
    )

    # Generate Platform-Specific requirements.txt
    if is_linux():
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "requirements.linux.txt"),
            os.path.join(os.getcwd(), "requirements.txt")
        )
    elif is_macos():
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "requirements.darwin.txt"),
            os.path.join(os.getcwd(), "requirements.txt")
        )
    elif is_windows():
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "requirements.win32.txt"),
            os.path.join(os.getcwd(), "requirements.txt")
        )

    # Install the dependencies
    print("Installing dependencies...")

    if is_linux():
        subprocess.call(shlex.split("pip install python-xlib"))
    elif is_macos():
        subprocess.call(shlex.split("pip install python-xlib pyobjc-framework-Quartz py-applescript"))
    elif is_windows():
        # Anaconda Packages
        subprocess.call(shlex.split("conda install numpy scipy scikit-image scikit-learn h5py -y"), shell=True)

    subprocess.call(shlex.split("pip install -r requirements.txt"))

    # Create Dataset Directories
    os.makedirs(os.path.join(os.getcwd(), "datasets/collect_frames"), exist_ok=True)
    os.makedirs(os.path.join(os.getcwd(), "datasets/collect_frames_for_context"), exist_ok=True)
    os.makedirs(os.path.join(os.getcwd(), "datasets/current"), exist_ok=True)


def setup_ocr():
    if is_linux():
        print("Before continuing with the OCR module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/Linux-Installation-Guide#ocr")
    elif is_macos():
        print("Before continuing with the OCR module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/macOS-Installation-Guide#ocr")
    elif is_windows():
        print("Before continuing with the OCR module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/Windows-Installation-Guide#ocr")

    print("")
    input("Press Enter to continue...")

    if is_unix():
        subprocess.call(shlex.split("pip install tesserocr"))
    elif is_windows():
        subprocess.call(shlex.split("pip install pytesseract"))

    print("")
    print("OCR module setup complete!")


def setup_gui():
    if is_linux():
        print("Before continuing with the GUI module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/Linux-Installation-Guide#gui")
    elif is_macos():
        print("Before continuing with the GUI module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/macOS-Installation-Guide#gui")
    elif is_windows():
        print("Before continuing with the GUI module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/Windows-Installation-Guide#gui")

    print("")
    input("Press Enter to continue...")

    if is_linux():
        subprocess.call(shlex.split("pip install Kivy"))
    elif is_macos():
        subprocess.call(shlex.split("pip install pygame Kivy"))
    elif is_windows():
        subprocess.call(shlex.split("pip install docutils pygments pypiwin32 kivy.deps.sdl2 kivy.deps.glew"))
        subprocess.call(shlex.split("pip install Kivy"))

    print("")
    print("GUI module setup complete!")


def setup_ml():
    if is_linux():
        print("Before continuing with the ML module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/Linux-Installation-Guide#ml")
    elif is_macos():
        print("Before continuing with the ML module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/macOS-Installation-Guide#ml")
    elif is_windows():
        print("Before continuing with the ML module setup, please read and perform the installation steps from the wiki: https://github.com/SerpentAI/SerpentAI/wiki/Windows-Installation-Guide#ml")

    print("")
    input("Press Enter to continue...")

    # Decide on CPU or GPU Tensorflow
    tensorflow_backend = input("\nWhich backend do you plan to use for Tensorflow (One of: 'CPU', 'GPU' - Note: GPU backend can only be used on NVIDIA GTX 600 series and up): \n")

    if tensorflow_backend not in ["CPU", "GPU"]:
        tensorflow_backend = "CPU"

    if tensorflow_backend == "GPU":
        subprocess.call(shlex.split("pip install tensorflow-gpu==1.4.1"))
    elif tensorflow_backend == "CPU":
        subprocess.call(shlex.split("pip install tensorflow==1.4.1"))

    subprocess.call(shlex.split("pip install Keras"))
    
    print("")
    print("ML module setup complete!")


def update():
    clear_terminal()
    display_serpent_logo()
    print("")

    print("Updating Serpent.AI to the latest version...")

    subprocess.call(shlex.split("pip install --upgrade SerpentAI"))

    if is_linux():
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "requirements.linux.txt"),
            os.path.join(os.getcwd(), "requirements.txt")
        )
    elif is_macos():
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "requirements.darwin.txt"),
            os.path.join(os.getcwd(), "requirements.txt")
        )
    elif is_windows():
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "requirements.win32.txt"),
            os.path.join(os.getcwd(), "requirements.txt")
        )

    subprocess.call(shlex.split("pip install -r requirements.txt"))

    import yaml

    with open(os.path.join(os.path.dirname(__file__), "config", "config.yml"), "r") as f:
        serpent_config = yaml.safe_load(f) or {}

    with open(os.path.join(os.getcwd(), "config", "config.yml"), "r") as f:
        user_config = yaml.safe_load(f) or {}

    config_changed = False

    for key, value in serpent_config.items():
        if key not in user_config:
            user_config[key] = value
            config_changed = True

    if config_changed:
       with open(os.path.join(os.getcwd(), "config", "config.yml"), "w") as f:
           f.write(yaml.dump(user_config)) 


def modules():
    import importlib
    exists = importlib.util.find_spec

    serpent_modules = {
        "OCR": (exists("tesserocr") or exists("pytesseract")) is not None,
        "GUI": exists("kivy") is not None,
        "ML": exists("keras") is not None
    }

    clear_terminal()
    display_serpent_logo()
    print("")

    print("Installed Serpent.AI Modules:")
    print("")

    print(f"OCR => {'Yes' if serpent_modules['OCR'] else 'No; Install with `serpent setup ocr` if needed'}")
    print(f"GUI => {'Yes' if serpent_modules['GUI'] else 'No; Install with `serpent setup gui` if needed'}")
    print(f"ML => {'Yes' if serpent_modules['ML'] else 'No; Install with `serpent setup ml` if needed'}")

    print("")


def grab_frames(width, height, x_offset, y_offset, pipeline_string=None):
    from serpent.frame_grabber import FrameGrabber

    frame_grabber = FrameGrabber(
        width=int(width),
        height=int(height),
        x_offset=int(x_offset),
        y_offset=int(y_offset),
        pipeline_string=pipeline_string
    )

    frame_grabber.start()


def activate(plugin_name):
    subprocess.call(shlex.split(f"offshoot install {plugin_name}"))


def deactivate(plugin_name):
    subprocess.call(shlex.split(f"offshoot uninstall {plugin_name}"))


def plugins():
    plugin_names = set()

    for root, directories, files in os.walk(offshoot.config['file_paths']['plugins']):
        if root != offshoot.config['file_paths']['plugins']:
            break

        for directory in directories:
            plugin_names.add(directory)

    manifest_plugin_names = set()

    for plugin_name in offshoot.Manifest().list_plugins().keys():
        manifest_plugin_names.add(plugin_name)

    active_plugins = plugin_names & manifest_plugin_names
    inactive_plugins = plugin_names - manifest_plugin_names

    print("\nACTIVE Plugins:\n")
    print("\n".join(active_plugins or ["No active plugins..."]))

    print("\nINACTIVE Plugins:\n")
    print("\n".join(inactive_plugins or ["No inactive plugins..."]))


def launch(game_name):
    game = initialize_game(game_name)
    game.launch()


def play(game_name, game_agent_name, frame_handler=None):
    game = initialize_game(game_name)
    game.launch(dry_run=True)

    game_agent_class_mapping = offshoot.discover("GameAgent", selection=game_agent_name)
    game_agent_class = game_agent_class_mapping.get(game_agent_name)

    if game_agent_class is None:
        raise Exception(f"Game Agent '{game_agent_name}' wasn't found. Make sure the plugin is installed.")

    game.play(game_agent_class_name=game_agent_name, frame_handler=frame_handler)


def record(game_name, game_agent_name, frame_count=4, frame_spacing=4):
    game = initialize_game(game_name)
    game.launch(dry_run=True)

    game.play(
        game_agent_class_name=game_agent_name, 
        frame_handler="RECORD", 
        frame_count=int(frame_count),
        frame_spacing=int(frame_spacing)
    )


def generate(plugin_type):
    if plugin_type == "game":
        generate_game_plugin()
    elif plugin_type == "game_agent":
        generate_game_agent_plugin()
    else:
        raise Exception(f"'{plugin_type}' is not a valid plugin type...")


def train(training_type, *args):
    if training_type == "context":
        train_context(*args)


def capture(capture_type, game_name, interval=1, extra=None, extra_2=None):
    game = initialize_game(game_name)
    game.launch(dry_run=True)

    if capture_type not in ["frame", "context", "region"]:
        raise Exception("Invalid capture command.")

    if capture_type == "frame":
        game.play(frame_handler="COLLECT_FRAMES", interval=float(interval))
    elif capture_type == "context":
        game.play(frame_handler="COLLECT_FRAMES_FOR_CONTEXT", interval=float(interval), context=extra, screen_region=extra_2)
    elif capture_type == "region":
        game.play(frame_handler="COLLECT_FRAME_REGIONS", interval=float(interval), region=extra)


def visual_debugger(*buckets):
    from serpent.visual_debugger.visual_debugger_app import VisualDebuggerApp
    VisualDebuggerApp(buckets=buckets or None).run()


def window_name():
    clear_terminal()
    print("Open the Game manually.")

    input("\nPress Enter and then focus the game window...")

    window_controller = WindowController()

    time.sleep(5)

    focused_window_name = window_controller.get_focused_window_name()

    print(f"\nGame Window Detected! Please set the kwargs['window_name'] value in the Game plugin to:")
    print("\n" + focused_window_name + "\n")


def record_inputs():
    from serpent.input_recorder import InputRecorder

    input_recorder = InputRecorder()

    input_recorder.start()


def generate_game_plugin():
    clear_terminal()
    display_serpent_logo()
    print("")

    game_name = input("What is the name of the game? (Titleized, No Spaces i.e. AwesomeGame): \n")
    game_platform = input("How is the game launched? (One of: 'steam', 'executable', 'web_browser'): \n")

    if game_name in [None, ""]:
        raise Exception("Invalid game name.")

    if game_platform not in ["steam", "executable", "web_browser"]:
        raise Exception("Invalid game platform.")

    prepare_game_plugin(game_name, game_platform)

    subprocess.call(shlex.split(f"serpent activate Serpent{game_name}GamePlugin"))


def generate_game_agent_plugin():
    clear_terminal()
    display_serpent_logo()
    print("")

    game_agent_name = input("What is the name of the game agent? (Titleized, No Spaces i.e. AwesomeGameAgent): \n")

    if game_agent_name in [None, ""]:
        raise Exception("Invalid game agent name.")

    prepare_game_agent_plugin(game_agent_name)

    subprocess.call(shlex.split(f"serpent activate Serpent{game_agent_name}GameAgentPlugin"))


def prepare_game_plugin(game_name, game_platform):
    plugin_destination_path = f"{offshoot.config['file_paths']['plugins']}/Serpent{game_name}GamePlugin".replace("/", os.sep)

    shutil.copytree(
        os.path.join(os.path.dirname(__file__), "templates/SerpentGamePlugin".replace("/", os.sep)),
        plugin_destination_path
    )

    # Plugin Definition
    with open(f"{plugin_destination_path}/plugin.py".replace("/", os.sep), "r") as f:
        contents = f.read()

    contents = contents.replace("SerpentGamePlugin", f"Serpent{game_name}GamePlugin")
    contents = contents.replace("serpent_game.py", f"serpent_{game_name}_game.py")

    with open(f"{plugin_destination_path}/plugin.py".replace("/", os.sep), "w") as f:
        f.write(contents)

    shutil.move(f"{plugin_destination_path}/files/serpent_game.py".replace("/", os.sep), f"{plugin_destination_path}/files/serpent_{game_name}_game.py".replace("/", os.sep))

    # Game
    with open(f"{plugin_destination_path}/files/serpent_{game_name}_game.py".replace("/", os.sep), "r") as f:
        contents = f.read()

    contents = contents.replace("SerpentGame", f"Serpent{game_name}Game")
    contents = contents.replace("MyGameAPI", f"{game_name}API")

    if game_platform == "steam":
        contents = contents.replace("PLATFORM", "steam")

        contents = contents.replace("from serpent.game_launchers.web_browser_game_launcher import WebBrowser", "")
        contents = contents.replace('kwargs["executable_path"] = "EXECUTABLE_PATH"', "")
        contents = contents.replace('kwargs["url"] = "URL"', "")
        contents = contents.replace('kwargs["browser"] = WebBrowser.DEFAULT', "")
    elif game_platform == "executable":
        contents = contents.replace("PLATFORM", "executable")

        contents = contents.replace("from serpent.game_launchers.web_browser_game_launcher import WebBrowser", "")
        contents = contents.replace('kwargs["app_id"] = "APP_ID"', "")
        contents = contents.replace('kwargs["app_args"] = None', "")
        contents = contents.replace('kwargs["url"] = "URL"', "")
        contents = contents.replace('kwargs["browser"] = WebBrowser.DEFAULT', "")
    elif game_platform == "web_browser":
        contents = contents.replace("PLATFORM", "web_browser")

        contents = contents.replace('kwargs["app_id"] = "APP_ID"', "")
        contents = contents.replace('kwargs["app_args"] = None', "")
        contents = contents.replace('kwargs["executable_path"] = "EXECUTABLE_PATH"', "")

    with open(f"{plugin_destination_path}/files/serpent_{game_name}_game.py".replace("/", os.sep), "w") as f:
        f.write(contents)

    # Game API
    with open(f"{plugin_destination_path}/files/api/api.py".replace("/", os.sep), "r") as f:
        contents = f.read()

    contents = contents.replace("MyGameAPI", f"{game_name}API")

    with open(f"{plugin_destination_path}/files/api/api.py".replace("/", os.sep), "w") as f:
        f.write(contents)


def prepare_game_agent_plugin(game_agent_name):
    plugin_destination_path = f"{offshoot.config['file_paths']['plugins']}/Serpent{game_agent_name}GameAgentPlugin".replace("/", os.sep)

    shutil.copytree(
        os.path.join(os.path.dirname(__file__), "templates/SerpentGameAgentPlugin".replace("/", os.sep)),
        plugin_destination_path
    )

    with open(f"{plugin_destination_path}/plugin.py", "r") as f:
        contents = f.read()

    contents = contents.replace("SerpentGameAgentPlugin", f"Serpent{game_agent_name}GameAgentPlugin")
    contents = contents.replace("serpent_game_agent.py", f"serpent_{game_agent_name}_game_agent.py")

    with open(f"{plugin_destination_path}/plugin.py", "w") as f:
        f.write(contents)

    shutil.move(f"{plugin_destination_path}/files/serpent_game_agent.py", f"{plugin_destination_path}/files/serpent_{game_agent_name}_game_agent.py")

    with open(f"{plugin_destination_path}/files/serpent_{game_agent_name}_game_agent.py", "r") as f:
        contents = f.read()

    contents = contents.replace("SerpentGameAgent", f"Serpent{game_agent_name}GameAgent")

    with open(f"{plugin_destination_path}/files/serpent_{game_agent_name}_game_agent.py", "w") as f:
        f.write(contents)


def train_context(epochs=3, validate=True, autosave=False):
    if validate not in [True, "True", False, "False"]:
        raise ValueError("'validate' should be True or False")

    if autosave not in [True, "True", False, "False"]:
        raise ValueError("'autosave' should be True or False")

    from serpent.machine_learning.context_classification.context_classifier import ContextClassifier
    ContextClassifier.executable_train(epochs=int(epochs), validate=argv_is_true(validate), autosave=argv_is_true(autosave))


def initialize_game(game_name):
    game_class_name = f"Serpent{game_name}Game"

    game_class_mapping = offshoot.discover("Game")
    game_class = game_class_mapping.get(game_class_name)

    if game_class is None:
        raise Exception(f"Game '{game_name}' wasn't found. Make sure the plugin is installed.")

    game = game_class()

    return game


def argv_is_true(arg):
    return arg in [True, "True"]


command_function_mapping = {
    "setup": setup,
    "update": update,
    "modules": modules,
    "grab_frames": grab_frames,
    "activate": activate,
    "deactivate": deactivate,
    "plugins": plugins,
    "launch": launch,
    "play": play,
    "record": record,
    "generate": generate,
    "train": train,
    "capture": capture,
    "visual_debugger": visual_debugger,
    "window_name": window_name,
    "record_inputs": record_inputs
}

command_description_mapping = {
    "setup": "Perform first time setup for the framework",
    "update": "Update the framework to the latest version",
    "modules": "List the install status of the framework's optional modules",
    "grab_frames": "Start the frame grabber",
    "activate": "Activate a plugin",
    "deactivate": "Deactivate a plugin",
    "plugins": "List all locally-available plugins",
    "launch": "Launch a game through a plugin",
    "play": "Play a game with a game agent through plugins",
    "record": "Record player input from a game",
    "generate": "Generate code for game and game agent plugins",
    "train": "Train a context classifier with collected context frames",
    "capture": "Capture frames, screen regions and contexts from a game",
    "visual_debugger": "Launch the visual debugger",
    "window_name": "Launch a utility to find a game's window name",
    "record_inputs": "Start the input recorder"
}

if __name__ == "__main__":
    execute()



import numpy as np

import mss

from redis import StrictRedis

import skimage.transform
import skimage.util

from serpent.config import config

import time

from serpent.game_frame import GameFrame
from serpent.game_frame_buffer import GameFrameBuffer

from serpent.frame_transformation_pipeline import FrameTransformationPipeline

from serpent.utilities import is_macos


redis_client = StrictRedis(**config["redis"])


class FrameGrabber:

    def __init__(self, width=640, height=480, x_offset=0, y_offset=0, fps=30, pipeline_string=None, buffer_seconds=5):
        self.width = width
        self.height = height

        self.x_offset = x_offset
        self.y_offset = y_offset

        self.frame_time = 1 / fps
        self.frame_buffer_size = buffer_seconds * fps

        self.redis_client = redis_client
        self.screen_grabber = mss.mss()

        self.frame_transformation_pipeline = None

        if pipeline_string is not None and isinstance(pipeline_string, str):
            self.frame_transformation_pipeline = FrameTransformationPipeline(pipeline_string=pipeline_string)

        self.is_retina_display = False
        self.is_retina_display = self._perform_retina_display_check()

        # Clear any previously stored frames
        self.redis_client.delete(config["frame_grabber"]["redis_key"])
        self.redis_client.delete(config["frame_grabber"]["redis_key"] + "_PIPELINE")

    def start(self):
        while True:
            cycle_start = time.time()

            frame = self.grab_frame()

            if self.frame_transformation_pipeline is not None:
                frame_pipeline = self.frame_transformation_pipeline.transform(frame)
            else:
                frame_pipeline = frame

            frame_shape = str(frame.shape).replace("(", "").replace(")", "")
            frame_dtype = str(frame.dtype)

            frame_bytes = f"{cycle_start}~{frame_shape}~{frame_dtype}~".encode("utf-8") + frame.tobytes()

            self.redis_client.lpush(config["frame_grabber"]["redis_key"], frame_bytes)
            self.redis_client.ltrim(config["frame_grabber"]["redis_key"], 0, self.frame_buffer_size)

            if self._has_png_transformation_pipeline():
                frame_pipeline_shape = "PNG"
                frame_pipeline_dtype = "PNG"

                frame_pipeline_bytes = f"{cycle_start}~{frame_pipeline_shape}~{frame_pipeline_dtype}~".encode("utf-8") + frame_pipeline
            else:
                frame_pipeline_shape = str(frame_pipeline.shape).replace("(", "").replace(")", "")
                frame_pipeline_dtype = str(frame_pipeline.dtype)

                frame_pipeline_bytes = f"{cycle_start}~{frame_pipeline_shape}~{frame_pipeline_dtype}~".encode("utf-8") + frame_pipeline.tobytes()

            self.redis_client.lpush(config["frame_grabber"]["redis_key"] + "_PIPELINE", frame_pipeline_bytes)
            self.redis_client.ltrim(config["frame_grabber"]["redis_key"] + "_PIPELINE", 0, self.frame_buffer_size)

            cycle_end = time.time()

            cycle_duration = (cycle_end - cycle_start)
            cycle_duration -= int(cycle_duration)

            frame_time_left = self.frame_time - cycle_duration

            if frame_time_left > 0:
                time.sleep(frame_time_left)

    def grab_frame(self):
        frame = np.array(
            self.screen_grabber.grab({
                "top": self.y_offset,
                "left": self.x_offset,
                "width": self.width,
                "height": self.height
            }),
            dtype="uint8"
        )

        frame = frame[..., [2, 1, 0, 3]]

        if self.is_retina_display:
            frame = skimage.util.img_as_ubyte(skimage.transform.resize(frame, (frame.shape[0] // 2, frame.shape[1] // 2)))
            frame = frame[:self.height, :self.width, :]

        return frame[..., :3]

    def _perform_retina_display_check(self):
        retina_display = False

        if is_macos():
            frame = self.grab_frame()

            if frame.shape[0] > self.height:
                retina_display = True

        return retina_display

    def _has_png_transformation_pipeline(self):
        return self.frame_transformation_pipeline and self.frame_transformation_pipeline.pipeline_string and self.frame_transformation_pipeline.pipeline_string.endswith("|PNG")

    @classmethod
    def get_frames(cls, frame_buffer_indices, frame_type="FULL", **kwargs):
        while True:
            if redis_client.llen(config["frame_grabber"]["redis_key"]) > 149:
                break
            
            time.sleep(0.1)

        game_frame_buffer = GameFrameBuffer(size=len(frame_buffer_indices))

        for i in frame_buffer_indices:
            redis_key = config["frame_grabber"]["redis_key"]
            redis_key = redis_key + "_PIPELINE" if frame_type == "PIPELINE" else redis_key

            frame_data = redis_client.lindex(redis_key, i)

            timestamp, shape, dtype, frame_bytes = frame_data.split("~".encode("utf-8"), maxsplit=3)

            if dtype == "PNG".encode("utf-8"):
                frame_array = frame_bytes
            else:
                frame_shape = [int(i) for i in shape.decode("utf-8").split(", ")]
                frame_array = np.fromstring(frame_bytes, dtype=dtype.decode("utf-8")).reshape(frame_shape)

            game_frame = GameFrame(frame_array, timestamp=float(timestamp))

            game_frame_buffer.add_game_frame(game_frame)

        return game_frame_buffer



#!/usr/bin/env python
from setuptools import setup

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = ""

packages = [
    "serpent",
    "serpent.game_agents",
    "serpent.game_launchers",
    "serpent.games",
    "serpent.input_controllers",
    "serpent.machine_learning",
    "serpent.machine_learning.context_classification",
    "serpent.machine_learning.context_classification.context_classifiers",
    "serpent.machine_learning.reinforcement_learning",
    "serpent.visual_debugger",
    "serpent.wamp_components",
    "serpent.window_controllers"
]

data_files = [
    ("serpent/templates/SerpentGameAgentPlugin", [
        "serpent/templates/SerpentGameAgentPlugin/plugin.py",
        "serpent/templates/SerpentGameAgentPlugin/__init__.py",
        "serpent/templates/SerpentGameAgentPlugin/.gitignore",
        "serpent/templates/SerpentGameAgentPlugin/.gitattributes"
    ]),
    ("serpent/templates/SerpentGameAgentPlugin/files", [
        "serpent/templates/SerpentGameAgentPlugin/files/serpent_game_agent.py",
        "serpent/templates/SerpentGameAgentPlugin/files/__init__.py"
    ]),
    ("serpent/templates/SerpentGameAgentPlugin/files/ml_models", [
        "serpent/templates/SerpentGameAgentPlugin/files/ml_models/.gitkeep"
    ]),
    ("serpent/templates/SerpentGameAgentPlugin/files/helpers", [
        "serpent/templates/SerpentGameAgentPlugin/files/helpers/.gitkeep"
    ]),
    ("serpent/templates/SerpentGamePlugin", [
        "serpent/templates/SerpentGamePlugin/plugin.py",
        "serpent/templates/SerpentGamePlugin/__init__.py",
        "serpent/templates/SerpentGamePlugin/.gitignore"
    ]),
    ("serpent/templates/SerpentGamePlugin/files", [
        "serpent/templates/SerpentGamePlugin/files/serpent_game.py",
        "serpent/templates/SerpentGamePlugin/files/__init__.py"
    ]),
    ("serpent/templates/SerpentGamePlugin/files/data/sprites", [
        "serpent/templates/SerpentGamePlugin/files/data/sprites/.gitkeep"
    ]),
    ("serpent/templates/SerpentGamePlugin/files/api", [
        "serpent/templates/SerpentGamePlugin/files/api/api.py",
        "serpent/templates/SerpentGamePlugin/files/api/__init__.py"
    ]),
    ("serpent/config", [
        "serpent/config/config.yml",
        "serpent/config/config.plugins.yml"
    ]),
    ("serpent", [
        "serpent/offshoot.manifest.json",
        "serpent/offshoot.yml",
        "serpent/requirements.linux.txt",
        "serpent/requirements.win32.txt",
        "serpent/requirements.darwin.txt"
    ])
]

requires = [
    "offshoot",
    "Cython==0.26.1"
]

setup(
    name='SerpentAI',
    version="2018.1.1",
    description='Game Agent Development Kit. Helping you create AIs / Bots to play any game you own!',
    long_description=long_description,
    author="Nicholas Brochu",
    author_email='nicholas@serpent.ai',
    packages=packages,
    data_files=data_files,
    include_package_data=True,
    install_requires=requires,
    entry_points={
        'console_scripts': ['serpent = serpent.serpent:execute']
    },
    license='MIT',
    url='https://github.com/SerpentAI/Serpent',
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Games/Entertainment'
    ]
)



from serpent.game_agent import GameAgent


class SerpentGameAgent(GameAgent):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.frame_handlers["PLAY"] = self.handle_play

        self.frame_handler_setups["PLAY"] = self.setup_play

        self.analytics_client = None

    def setup_play(self):
        pass

    def handle_play(self, game_frame):
        pass



from serpent.frame_transformer import FrameTransformer


class FrameTransformationPipelineError(BaseException):
    pass


class FrameTransformationPipeline:

    def __init__(self, pipeline_string=None):
        if pipeline_string is None or not isinstance(pipeline_string, str):
            raise FrameTransformationPipelineError("A 'pipeline_string' kwarg is expected...")

        self.game_frame_transformer = FrameTransformer()
        
        self.pipeline = self._parse_pipeline_string(pipeline_string)
        self.pipeline_string = pipeline_string

    @property
    def pipeline_operations(self):
        return {
            "RESIZE": self.game_frame_transformer.resize,
            "RESCALE": self.game_frame_transformer.rescale,
            "GRAYSCALE": self.game_frame_transformer.grayscale,
            "FLOAT": self.game_frame_transformer.to_float,
            "PNG": self.game_frame_transformer.to_png
        }

    def transform(self, frame=None):
        for pipeline_func, args in self.pipeline:
            if args[0] != "":
                frame = pipeline_func(frame, *args)
            else:
                frame = pipeline_func(frame)

        return frame

    def _parse_pipeline_string(self, pipeline_string):
        pipeline = list()

        string_pipeline_operations = pipeline_string.split("|")

        for string_pipeline_operation in string_pipeline_operations:
            if ":" not in string_pipeline_operation:
                string_pipeline_operation += ":"

            operation_key, args = string_pipeline_operation.split(":")

            if operation_key in self.pipeline_operations:
                pipeline.append((self.pipeline_operations[operation_key], args.split(",")))

        return pipeline



import skimage.transform
import skimage.color
import skimage.util

from PIL import Image

import io


class FrameTransformer:

    def __init__(self):
        pass

    @staticmethod
    def resize(frame, size_string=None):
        width, height = size_string.lower().split("x")
        return skimage.util.img_as_ubyte(skimage.transform.resize(frame, (int(height), int(width)), order=0))

    @staticmethod
    def rescale(frame, scale=None):
        return skimage.util.img_as_ubyte(skimage.transform.rescale(frame, float(scale)))

    @staticmethod
    def grayscale(frame):
        return skimage.util.img_as_ubyte(skimage.color.rgb2gray(frame))

    @staticmethod
    def to_float(frame):
        return skimage.util.img_as_float(frame)
    
    @staticmethod
    def to_png(frame):
        pil_frame = Image.fromarray(skimage.util.img_as_ubyte(frame))
        pil_frame = pil_frame.convert("RGB")

        png_frame = io.BytesIO()

        pil_frame.save(png_frame, format="PNG", compress_level=3)
        png_frame.seek(0)

        return png_frame.read()
